#include <cstdio>
#include <map>
#include <set>

using namespace std;

#include "instrument.hh"
#include "analysis.hh"
#include "snd.hh"

double TickSpacer;

static vector< set<unsigned> > BeatList;

static bool IsTickPoint(const SNDChannel *prev, const SNDChannel *now)
{
    double wlendiff;
    if(!prev)return now;
    
    if(now->wavelen == 0 || prev->wavelen == 0)
        wlendiff = 0;
    else if(now->wavelen > prev->wavelen)
        wlendiff = now->wavelen / (double) prev->wavelen;
    else
        wlendiff = prev->wavelen / (double) now->wavelen;
    
    if(!prev->vol
    || now->vol >= prev->vol+4
    || (prev->vol > 1 && now->vol >= prev->vol*3)
    || (now->type != SNDChannel::rnd && wlendiff >=
      // 1.0567
      // 1.0566
         1.0562
     )
       )
    {
        return true;
    }
    return false;
}

static unsigned GCD2(unsigned a, unsigned b)
{
    while(a > 0) { unsigned tmp = b % a; b = a; a = tmp; }
    return b;
}
static unsigned GCD(const set<unsigned> &nums)
{
    unsigned last = 0;
    for(set<unsigned>::const_iterator i = nums.begin(); i != nums.end(); ++i)
        last = last ? GCD2(*i, last) : *i;
    return last;
}

static unsigned LCD(const set<unsigned> &nums)
{
    unsigned a=1;
    for(set<unsigned>::const_iterator i = nums.begin(); i != nums.end(); ++i)
        a *= *i;
    return a;// / GCD(nums);
}

void AnalyzeSND(song &song)
{
    /* For DPCM channels the DPCM equipment only
     * takes care of instruments and they are rarely
     * tweaked, so we can take them just as raw */
     
    /* Often, it seems, samples start on one channel just a tick
     * before/after other channels. In these cases, it's considerable
     * to include that void tick in the sample.
     * Another way could be to move the sample backwards/forwards
     * for one tick measure...
     */

    vector<unsigned> lasttick(song.channelcount);
    for(unsigned c=0; c<song.channelcount; ++c)
        lasttick[c] = 0;
    vector<sample> cursample(song.channelcount);

    vector< vector<bool> > ticklists(song.channelcount);
    for(unsigned a=0; a<song.channelcount; ++a)
        ticklists[a].resize(song.rownum, false);
    
    BeatList.clear();
    BeatList.resize(song.channelcount);
    
    for(unsigned a=0; a<song.rownum; ++a)
    {
        for(unsigned c=0; c<song.channelcount; ++c)
        {
            vector<bool> &ticks = ticklists[c];
            const SNDChannel &tmp = song.ChanPtr(a, c);
            if(tmp.vol)
            {
                const SNDChannel *prev = a > 0 ? &song.ChanPtr(a-1, c) : NULL;
                
                sample &chansample = cursample[c];
                
                if(IsTickPoint(prev, &tmp))
                {
                    if(a != lasttick[c] + 1 || a == 1)
                    {
                        ticks[a] = true;
                        BeatList[c].insert(a);

                        if(tmp.type == SNDChannel::pcm)
                        {
                            continue;
                        }
            
                        if(chansample.alkutree.size())
                        {
                            unsigned ins = AnalyzeSample(chansample);
                            song.chanptr[chansample.beginrow * song.channelcount + c].ins = ins;
                        }
                        /* Start of a new sample! */
                        chansample.type        = tmp.type;
                        chansample.beginrow    = a;
                        chansample.basewavelen = tmp.wavelen;
                        chansample.basevol     = tmp.vol;
                        chansample.alkutree.clear();
                        /*
                        chansample.lopputree.clear();
                        */
                    }
                    lasttick[c] = a;
                }

                if(tmp.type == SNDChannel::pcm)
                {
                    continue;
                }
            
                if(chansample.basewavelen)
                {
                    sampleitem item;
                    item.duty       = tmp.sample;
                    item.wavelenadd = tmp.wavelen - chansample.basewavelen;
                    item.wavelenmul = tmp.wavelen * 2 / chansample.basewavelen;
                    item.voladd     = tmp.vol - chansample.basevol;
                    item.volmul     = tmp.vol * 2 / chansample.basevol;
                    
                    if(chansample.alkutree.size() < 15)
                    {
                        chansample.alkutree.push_back(item);
                    }
                    /*
                    chansample.lopputree.push_front(item);
                    if(chansample.lopputree.size() > chansample.alkutree.size())
                    {
                        chansample.lopputree.pop_back();
                    }
                    */
                }
            }
        }
    }
    
    map<unsigned, unsigned> lencounts;
    
    for(unsigned b=0; b<ticklists.size(); ++b)
    {
        vector<bool> &ticks = ticklists[b];
        unsigned a, firsttick;
        for(a=0; a<ticks.size() && !ticks[a]; ++a);
        firsttick=a;
        while(++a<ticks.size())
        {
            if(ticks[a]) { ++lencounts[a-firsttick]; firsttick=a; }
        }
    }
    unsigned mostticks = 0;
    for(map<unsigned,unsigned>::const_iterator
        i=lencounts.begin(); i!=lencounts.end(); ++i)
        if(i->second > mostticks)
            mostticks = i->second;

    for(map<unsigned,unsigned>::const_iterator
        i=lencounts.begin(); i!=lencounts.end(); ++i)
        fprintf(stderr, "%u:%u ", i->first, i->second);
    fprintf(stderr, "\n");

    set<double> lens;
    set<unsigned> divisors;

    for(map<unsigned,unsigned>::const_iterator
        i=lencounts.begin(); i!=lencounts.end(); ++i)
    {
        unsigned len = i->first;
        unsigned count=i->second;
        
        if(count < mostticks/20) continue;
        
        map<unsigned,unsigned>::const_iterator j = i; ++j;
        if(j != lencounts.end() && j->first == len+1)
        {
            unsigned len2 = j->first;
            unsigned count2 = j->second;
            
            if(count2 >= mostticks/20)
            {
                unsigned maara = count + count2;
                
                double suhde = (count > count2) ? count / (double)count2 : count2 / (double)count;
                if(suhde < 30)
                {
                    double uusi = (len * count + len2 * count2) / (double)maara;
                    i = j;
                    lens.insert(uusi);
                    divisors.insert(maara);
                    continue;
                }
            }
        }

        lens.insert(len);
    }

    for(set<double>::const_iterator
        i=lens.begin(); i!=lens.end(); ++i)
        fprintf(stderr, "%g ", *i);
    fprintf(stderr, "\n");

    unsigned divisor = LCD(divisors);
    set<unsigned> intlens;
    for(set<double>::const_iterator i = lens.begin(); i != lens.end(); ++i)
        intlens.insert((unsigned) (*i * divisor));
    double step = GCD(intlens) / (double)divisor;
    
    if(step < 1.0 && ((1.0 / step) - (unsigned)(1.0 / step)) < 0.2)
        step = 1.0;
    fprintf(stderr, "Step: %g\n", step);
    
    while(step >= 4.0) step /= 2;
    TickSpacer = step;
}

bool IsBeat(unsigned row, unsigned chan)
{
    return BeatList[chan].find(row) != BeatList[chan].end();
}
